/*
SDX: Documentary System in XML.
Copyright (C) 2000, 2001, 2002  Ministere de la culture et de la communication (France), AJLSM

Ministere de la culture et de la communication,
Mission de la recherche et de la technologie
3 rue de Valois, 75042 Paris Cedex 01 (France)
mrt@culture.fr, michel.bottin@culture.fr

AJLSM, 17, rue Vital Carles, 33000 Bordeaux (France)
sevigny@ajlsm.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the
Free Software Foundation, Inc.
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
or connect to:
http://www.fsf.org/copyleft/gpl.html
*/
package fr.gouv.culture.sdx.exception;

import fr.gouv.culture.sdx.framework.Framework;
import fr.gouv.culture.sdx.utils.Utilities;
import fr.gouv.culture.sdx.utils.constants.Node;
import org.apache.avalon.framework.logger.Logger;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.xml.XMLizable;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.MessageFormat;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

/**
 * The superclass of all exceptions thrown by the SDX Framework and the applications contained within.
 */
public class SDXException extends Exception implements XMLizable, SDXExceptionCode {

    /** The current locale (system's locale by default). */
    protected Locale locale = Locale.getDefault();

    /** The base name for the property file. */
    private final static String propertiesBaseName = "fr.gouv.culture.sdx.exception.errors";

    /** The error code. */
    private int errorCode = -1;

    /** The error key for the properties files. */
    private String errorKey;

    /** The error arguments. */
    private String[] arguments;

    /** The logger to use. */
    private Logger logger = null;

    /** The original exception. */
    private Throwable originalError = null;

    /**Our default locale for resource loading if the error_locale.properties file doesn't exist for the current system locale*/
    private Locale defaultLocale = Locale.ENGLISH;


    //TODO: another constructor accepting a string (e.getMessage()) and and exception so that sdx developers can have a convient way to build an SDXE-rbp
    /**
     * Creates an SDX exception with all parameters.
     *
     * @param l         The logger to use (may be null).
     * @param code      The error code (must be defined in class SDXExceptionCode).
     * @param args      The message arguments (may be null or empty).
     * @param e         The original exception thrown (may be null).
     */
    public SDXException(Logger l, int code, String[] args, Throwable e) {
        this.logger = l;
        this.errorCode = code;
        this.errorKey = "error." + this.errorCode;
        this.originalError = e;
        if (args != null)
            this.arguments = args;
        else
            this.arguments = new String[0];
        //we dont want any nulls sent to the message formatting methods
        for (int i = 0; i < this.arguments.length; i++) {
            if (this.arguments[i] == null)
                this.arguments[i] = "missing exception detail";
        }
        log();
    }


    /** Logs an error message. */
    private void log() {
        if (logger != null) {
            String message = getMessage();
            if (!Utilities.checkString(message))
                message = "noMessageProvided";

            if (originalError != null && originalError.fillInStackTrace() != null)
                logger.error(message, originalError.fillInStackTrace());
            else
                logger.error(message, this);
        }
        //a null logger allows  us to wrap our exceptions within other types of exceptions without double logging (example an Avalon ConfigurationException)
    }

    /** Returns an XML description of this error.TODOJavadoc:once we have a schema for the structure of this representation we should link to it here */
    public void toSAX(ContentHandler hdl) throws SAXException, ProcessingException {
        toSAX(hdl, this.locale);
    }

    private void toSAX(ContentHandler hdl, Locale locale) throws SAXException {
        // TODOImplement: Implement multilinguism.
        String sdxNsUri = Framework.SDXNamespaceURI;
        String sdxNsPrefix = Framework.SDXNamespacePrefix;

        //Creation of local variables which are later passed into startElement() and endElement() methods-rbp14/03/02
        String localName = Node.Name.EXCEPTION;
        String qName = sdxNsPrefix + ":" + localName;
        AttributesImpl atts = new AttributesImpl();
        String lang = locale.getLanguage();
        if (!"".equals(lang) && !"".equals(locale.getCountry())) lang += "-" + locale.getCountry();
        atts.addAttribute(Framework.XMLNamespaceURI, Node.Name.LANG, Node.Name.XML_LANG, Node.Type.CDATA, lang);
        atts.addAttribute("", Node.Name.CODE, Node.Name.CODE, Node.Type.CDATA, Integer.toString(this.errorCode));

        //startElement() method is called for "exception" and local variables are passed-rbp14/03/02
        hdl.startElement(sdxNsUri, localName, qName, atts);


        String childTwoLocalName = Node.Name.MESSAGE;
        String childTwoQName = sdxNsPrefix + ":" + childTwoLocalName;
        AttributesImpl childTwoAtts = new AttributesImpl();

        //startElement() method is called for "message" and local variables are passed-rbp14/03/02
        hdl.startElement(sdxNsUri, childTwoLocalName, childTwoQName, childTwoAtts);

        //Creation of an characterArray to input pass into characters() method-rbp14/03/02
        char[] messArr = this.getMessage(locale).toCharArray();
        hdl.characters(messArr, 0, messArr.length);

        //endElement() method is called for "message" and local variables are passed-rbp14/03/02
        hdl.endElement(sdxNsUri, childTwoLocalName, childTwoQName);

        if (originalError != null) {
            String childThreeLocalName = Node.Name.ORIGINAL_EXCEPTION;
            String childThreeQName = sdxNsPrefix + ":" + childThreeLocalName;
            AttributesImpl childThreeAtts = new AttributesImpl();



            //startElement() method is called for "originalError" and local variables are passed-rbp14/03/02
            hdl.startElement(sdxNsUri, childThreeLocalName, childThreeQName, childThreeAtts);

            try {
                //in java1.4 the  below code will not be necessary
                ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
                PrintStream ps = new PrintStream(byteOut);
                originalError.printStackTrace(ps);
//              char[] origExceptArr = (originalError.getMessage()).toCharArray();
                //Creation of an characterArray to input pass into characters() method-rbp14/03/02
                byteOut.flush();
                char[] origExceptArr = (byteOut.toString()).toCharArray();
                byteOut.close();
                hdl.characters(origExceptArr, 0, origExceptArr.length);
            } catch (IOException e) {
                throw new SAXException(e.getMessage(), e);
            } finally {
                //endElement() method is called for "originalError" and local variables are passed-rbp14/03/02
                hdl.endElement(sdxNsUri, childThreeLocalName, childThreeQName);
            }

        }
        //endElement() method is called for "exception" and local variables are passed-rbp14/03/02
        hdl.endElement(sdxNsUri, localName, qName);

    }

    /**
     * Returns an error message.
     *
     * <p>
     * The error message is in the default locale of the Java virtual machine, if available, if not
     * in the SDX default (english)
     * <p>TODONOW:currently not true, the default locale (english) still needs to be implemented-rbp
     */
    public String getMessage() {
        return getMessage(locale);
    }

    /**
     * Returns the error message for a given locale.
     *
     * @param   l      The locale.
     */
    public String getMessage(Locale l) {
        if (l != null) locale = l;
        ResourceBundle rb = null;
        try {
            rb = ResourceBundle.getBundle(propertiesBaseName, locale);
        } catch (MissingResourceException e) {
            rb = ResourceBundle.getBundle(propertiesBaseName, defaultLocale);
        }
        return MessageFormat.format(rb.getString(errorKey), arguments);
    }


}
